dofile("util.lua")

Widget_pkg = {}

function Widget_pkg.shallow_copy_table(t) 
   local newt = {}
   for i,v in t do
      newt[i] = v
   end
   return newt
end

-- old_globals = Widget_pkg.shallow_copy_table(globals())

-- The structure of a Widget

Widget_pkg.Widget_tag = newtag()

-- A Widget is a proxy for an underlying Fl_Widget, given in
-- widget.fl_widget.  A pointer back to the Widget is stored in the
-- Fl_Widget's clone table:

function Widget_pkg.fl_widget_from_widget(w)
   return w.fl_widget
end

function Widget_pkg.widget_from_fl_widget(fl_w)
   return fl_w._widget
end

function Widget_pkg.is_fl_widget(o)
   -- uhhh...let's cheat for now
   return tag(o) >= tag(Fl_Widget)
end

function Widget_pkg.is_widget(o)
   return tag(o) == Widget_pkg.Widget_tag
end

-- A table, widget.__class, contains special handling instructions for
-- sets and gets.
-- 
-- Right now, Widget also stores labels in widget.__stashed_label to
-- protect them from garbage collection.

-- Given a return from an Fl_ function, translate it back into a
-- Widget if necessary.

function Widget_pkg.fl_return_value_translate(v)
   if Widget_pkg.is_fl_widget(v) then
      return Widget_pkg.widget_from_fl_widget(v)
   else
      return v
   end
end

-- Make a new list with all Widgets turned int Fl_Widgets.  Used for
-- arg lists.
function Widget_pkg.fl_widget_list_from_widget_list(fl_l)
   local n = {}
   for i,v in fl_l do
      if Widget_pkg.is_widget(v) then
         n[i] = Widget_pkg.fl_widget_from_widget(v)
      else
         n[i] = v
      end
   end
   return n
end

-- Classes currently contain two tables, class.get_attribute_map and
-- .set_attribute_map.  They map from widget field names to handling
-- instructions.

-- String-valued entries describe verbatim maps from attribute
-- gets/sets to fl_widget:attribute()/fl_widget:attribute(value) calls.

function Widget_pkg.add_identity_maps(table, names)
   for i=1,getn(names) do 
      table[names[i]] = {name=names[i]}
   end
end

-- Function-valued entries are handed the Widget, the Fl_Widget, and
-- the name of the field.  Setters get the target value as well.

-- Table-valued entries are of the form {name="fl_function-name",
-- remap=1} for now; they convert the set/get value apropriately
-- to/from Fl_Widgets.

function Widget_pkg.add_translate_maps(table, names)
   for i=1,getn(names) do 
      table[names[i]] = {name=names[i], remap=1}
   end
end

function Widget_pkg.widget_gettable(table, index) 
   local class = rawget(table, "__class")
   local fl_widget = rawget(table, "fl_widget")
   -- print("fl_widget", fl_widget)
   -- class._Widget_pkg.widget_gettable(table, index)
   local target = class.get_attribute_map[index]
   if target then
      -- print("got target")
      if (type(target) == "table") then
         local base_result = fl_widget[target.name](fl_widget)
         if target.remap then
            return Widget_pkg.widget_from_fl_widget(base_result)
         else
            return base_result
         end
      else -- should be a function
         return target(table, fl_widget, index)
      end
   else
      local local_value = rawget(table, index)
      if local_value then return local_value end
      -- forward to widget
      -- return fl_widget[index]
      -- actually we need to shim this
      local fl_function = fl_widget[index]
      -- print("func", fl_function)
      if fl_function then
         return function (...) 
                   local new_arg = Widget_pkg.fl_widget_list_from_widget_list(arg)
                   return Widget_pkg.fl_return_value_translate(call(%fl_function, new_arg))
                end
      else
         return nil
      end
   end
end

function Widget_pkg.widget_settable(table, index, value)
   local class = rawget(table, "__class")
   local fl_widget = rawget(table, "fl_widget")
   -- class._Widget_pkg.widget_gettable(table, index)
   local target = class.set_attribute_map[index]
   if target then
      if (type(target) == "table") then
         if (target.remap) then
            value = Widget_pkg.fl_widget_from_widget(value)
         end
         return fl_widget[target.name](fl_widget, value)
      else
         return target(table, fl_widget, index, value)
      end
   else
      rawset(table, index, value)
   end
end




-- If the widget has a callback, call it; if the fl_widget has one, do
-- it as well.
function fltk_dispatch(fl_w) 
   local w = fl_w._widget
   local target
   if w then
      if w.callback then
         w:callback()
      end
   end
   if fl_w.callback then
      fl_w:callback()
   end
end



settagmethod(Widget_pkg.Widget_tag, "gettable", Widget_pkg.widget_gettable)
settagmethod(Widget_pkg.Widget_tag, "settable", Widget_pkg.widget_settable)

function Widget_pkg.insert_parent_constructors(class, table)
   if class then
      if class.parent then
         Widget_pkg.insert_parent_constructors(class.parent, table)
      end
      if class.initializers then
         for i,v in class.initializers do
            table[i] = v
         end
      end
      if class.ignore_initializers then
         for i,v in class.ignore_initializers do
            table[i] = nil
         end
      end
   end
end

function Widget_pkg.make_constructor(fl_class, class_table)
   return function (t)
             local initializer_sum = {}
             Widget_pkg.insert_parent_constructors(%class_table, initializer_sum)
             local constructor_args = {%fl_class}
             for i = 1,4 do
                constructor_args[i+1] = t[i]
             end
             local fl_widget = call(%fl_class.new, constructor_args)
             local w = {__class=%class_table, fl_widget=fl_widget, __stashed_label=""}
             fl_widget._widget = w
             fl_widget:label(__stashed_label)
             settag(w, Widget_pkg.Widget_tag)
             for i,v in initializer_sum do
                w[i] = v
             end
             for i,v in t do
                if type(i) ~= "number" then
                   w[i] = v
                end
             end
             -- If there was a fifth constructor arg, it's the label.
             if t[5] then w.label = t[5] end
             return w
          end
end



function Widget_pkg.clone_class(c)
   return {
      get_attribute_map=Widget_pkg.shallow_copy_table(c.get_attribute_map),
      set_attribute_map=Widget_pkg.shallow_copy_table(c.set_attribute_map), 
      initializers={},
      ignore_initializers={},
      parent=c
   }
end

-------------------- Widget
Widget_class = {
   get_attribute_map={},
   set_attribute_map={},
   initializers={},
   ignore_initializers={},
   parent = nil
}

--#Widget_class = {
--   attribute_map = { align={name="align"}, 

Widget_pkg.add_identity_maps(Widget_class.get_attribute_map, 
  {"align", "box", "color", "selection_color", "labeltype", "labelfont", "labelsize",
     "when", "visible", "active", "output", "changed", "type"})

Widget_pkg.add_identity_maps(Widget_class.get_attribute_map, 
		  {"x", "y", "w", "h"})


Widget_pkg.add_identity_maps(Widget_class.set_attribute_map,
  {"align", "box", "color", "selection_color", "labeltype", "labelfont", "labelsize",
     "when", "type"})
-- visible, active, output, changed


Widget_pkg.add_translate_maps(Widget_class.get_attribute_map, {"window"})

Widget_class.get_attribute_map.label = function (table, fl_widget, index)
   return rawget(table, "__stashed_label")
end

Widget_class.set_attribute_map.label = function(table, fl_widget, index, value)
   rawset(table, "__stashed_label", value)
   fl_widget:label(value)
end

Widget_class.set_attribute_map.callback = function(table, fl_widget, index, value)
                                             rawset(table, index, value)
                                             fl_cb(fl_widget)
end

-------------------- Box
Box = Widget_pkg.make_constructor(Fl_Box, Widget_class)

-------------------- Browser_

Browser__class = Widget_pkg.clone_class(Widget_class)
Widget_pkg.add_identity_maps(Browser__class.get_attribute_map, 
                  {"position", "hposition", "has_scrollbar", "textfont", "textsize", "textcolor", "scrollbar_width"})


Widget_pkg.add_identity_maps(Browser__class.set_attribute_map,
                  {"position", "hposition", "has_scrollbar", "textfont", "textsize", "textcolor", "scrollbar_width"})

-------------------- Browser

Browser_class = Widget_pkg.clone_class(Browser__class)
Widget_pkg.add_identity_maps(Browser_class.get_attribute_map, 
                  {"size", "topline", "selection", "value", "format_char", "column_char"})

Widget_pkg.add_identity_maps(Browser_class.set_attribute_map, 
                  {"value", "format_char", "column_char", "topline"})

Browser = Widget_pkg.make_constructor(Fl_Browser, Browser_class)

-------------------- Hold_Browser

Hold_Browser = Widget_pkg.make_constructor(Fl_Hold_Browser, Browser_class)

-------------------- Multi_Browser

Multi_Browser = Widget_pkg.make_constructor(Fl_Multi_Browser, Browser_class)

-------------------- Select_Browser

Select_Browser = Widget_pkg.make_constructor(Fl_Select_Browser, Browser_class)

-------------------- Button
Button_class = Widget_pkg.clone_class(Widget_class)
Widget_pkg.add_identity_maps(Button_class.get_attribute_map, 
                  {"value", "shortcut", "down_box"})
Widget_pkg.add_identity_maps(Button_class.set_attribute_map, 
                  {"value", "shortcut", "down_box"})

Button = Widget_pkg.make_constructor(Fl_Button, Button_class)

-------------------- Check_Button 
Check_Button_class = Widget_pkg.clone_class(Button_class)

Check_Button = Widget_pkg.make_constructor(Fl_Check_Button, Check_Button_class)

-------------------- Light_Button 

Light_Button = Widget_pkg.make_constructor(Fl_Light_Button, Button_class)

-------------------- Repeat_Button 

Repeat_Button = Widget_pkg.make_constructor(Fl_Repeat_Button, Button_class)

-------------------- Return_Button 

Return_Button = Widget_pkg.make_constructor(Fl_Return_Button, Button_class)

-------------------- Round_Button 

Round_Button = Widget_pkg.make_constructor(Fl_Round_Button, Button_class)

-------------------- Group
Group_class = Widget_pkg.clone_class(Widget_class)
Widget_pkg.add_translate_maps(Group_class.get_attribute_map, {"resizable"})
Widget_pkg.add_translate_maps(Group_class.set_attribute_map, {"resizable"})

Group = Widget_pkg.make_constructor(Fl_Group, Group_class)

-------------------- Pack
Pack_class = Widget_pkg.clone_class(Group_class)
Widget_pkg.add_identity_maps(Pack_class.get_attribute_map, {"spacing"})
Widget_pkg.add_identity_maps(Pack_class.set_attribute_map, {"spacing"})

Pack = Widget_pkg.make_constructor(Fl_Pack, Pack_class)

-------------------- Scroll
Scroll_class = Widget_pkg.clone_class(Group_class)
Widget_pkg.add_identity_maps(Scroll_class.get_attribute_map, {"xposition", "yposition"})

Scroll = Widget_pkg.make_constructor(Fl_Scroll, Scroll_class)

-------------------- Tabs
Tabs_class = Widget_pkg.clone_class(Group_class)
Widget_pkg.add_translate_maps(Tabs_class.get_attribute_map, {"value"})
Widget_pkg.add_translate_maps(Tabs_class.set_attribute_map, {"value"})

Tabs = Widget_pkg.make_constructor(Fl_Tabs, Tabs_class)

-------------------- Tile
Tile = Widget_pkg.make_constructor(Fl_Tile, Group_class)

-------------------- Window
Window_class = Widget_pkg.clone_class(Group_class)
Widget_pkg.add_identity_maps(Window_class.get_attribute_map, 
                  {"border", "shown", "modal", "non_modal"})
Widget_pkg.add_identity_maps(Window_class.set_attribute_map, 
                  {"border"})

-- These two are questionable; does the library copy them?
Widget_pkg.add_identity_maps(Window_class.get_attribute_map, 
                  {"iconlabel", "xclass"})
Widget_pkg.add_identity_maps(Window_class.set_attribute_map, 
                  {"iconlabel", "xclass"})

Window = Widget_pkg.make_constructor(Fl_Window, Window_class)


-------------------- Input_
Input__class = Widget_pkg.clone_class(Widget_class)

Widget_pkg.add_identity_maps(Input__class.get_attribute_map, 
                  {"value", "size", "maximum_size", "position", "mark",
                     "textfont", "textsize", "textcolor", "cursor_color"})
Widget_pkg.add_identity_maps(Input__class.set_attribute_map, 
                  {"value","maximum_size", "position", "mark",
                     "textfont", "textsize", "textcolor", "cursor_color"})

-------------------- Input
Input = Widget_pkg.make_constructor(Fl_Input, Input__class)

-------------------- Float_Input
Float_Input = Widget_pkg.make_constructor(Fl_Float_Input, Input__class)

-------------------- Int_Input
Int_Input = Widget_pkg.make_constructor(Fl_Int_Input, Input__class)

-------------------- Multiline_Input
Multiline_Input = Widget_pkg.make_constructor(Fl_Multiline_Input, Input__class)

-------------------- Secret_Input
Secret_Input = Widget_pkg.make_constructor(Fl_Secret_Input, Input__class)

-------------------- Output
Output = Widget_pkg.make_constructor(Fl_Output, Input__class)

-------------------- Multiline_Output
Multiline_Output = Widget_pkg.make_constructor(Fl_Multiline_Output, Input__class)

-------------------- Menu_
Menu__class = Widget_pkg.clone_class(Widget_class)
Widget_pkg.add_identity_maps(Menu__class.get_attribute_map, 
                  {"size", "value", "text", "textfont", "textsize", "textcolor", "down_box"})
Widget_pkg.add_identity_maps(Menu__class.set_attribute_map, 
                  {"size", "value", "textfont", "textsize", "textcolor", "down_box"})

-------------------- Choice
Choice = Widget_pkg.make_constructor(Fl_Choice, Menu__class)

-------------------- Menu_Bar
Menu_Bar = Widget_pkg.make_constructor(Fl_Menu_Bar, Menu__class)

-------------------- Menu_Button
Menu_Button = Widget_pkg.make_constructor(Fl_Menu_Button, Menu__class)


-- for i,v in globals() do  if not old_globals[i] then print(i, " = ", v)  end end
